// Copyright 2000-2021 JetBrains s.r.o.
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
// https://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.
package com.goide.utils;

import org.jetbrains.annotations.Contract;
import org.jetbrains.annotations.NotNull;
import org.jetbrains.annotations.Nullable;

import java.util.Locale;

public class GoStringUtil {
  private static final int[] EMPTY_INT_ARRAY = new int[0];

  @NotNull
  @Contract(pure = true)
  public static String repeat(@NotNull String s, int count) {
    if (count == 0) return "";
    assert count >= 0 : count;
    StringBuilder sb = new StringBuilder(s.length() * count);
    for (int i = 0; i < count; i++) {
      sb.append(s);
    }
    return sb.toString();
  }

  @Nullable
  public static String unescapeStringLiteralText(@Nullable String stringLiteralText) {
    GoStringUtil.GoDecodedString decodedString = decodeStringLiteralExpression(stringLiteralText);
    return decodedString.correct ? decodedString.unescapedString : null;
  }

  public static GoStringUtil.GoDecodedString decodeStringLiteralExpression(@Nullable String stringLiteralText) {
    String unquoted = unquoteStringLiteralExpression(stringLiteralText);
    if (unquoted == null || unquoted.isEmpty()) {
      return new GoStringUtil.GoDecodedString(unquoted != null, EMPTY_INT_ARRAY, "");
    }
    boolean raw = stringLiteralText.charAt(0) == '`';
    return raw ? new GoStringUtil.GoDecodedString(true, EMPTY_INT_ARRAY, unquoted) : decode(unquoted);
  }

  @NotNull
  public static String createSingleLineString(@NotNull String stringContent) {
    StringBuilder result = new StringBuilder();
    result.append("\"");
    escapeString(stringContent, result);
    result.append("\"");
    return result.toString();
  }

  @NotNull
  public static String escapeString(@NotNull String string) {
    StringBuilder escapedString = new StringBuilder();
    escapeString(string, escapedString);
    return escapedString.toString();
  }

  /**
   * Escapes the specified string in accordance with https://golang.org/ref/spec#Rune_literals
   */
  public static void escapeString(@NotNull String chars, @NotNull StringBuilder outChars) {
    int index = 0;

    while (index < chars.length()) {
      int c = chars.codePointAt(index);

      switch (c) {
        case (char)7:
          outChars.append("\\a");
          break;

        case '\b':
          outChars.append("\\b");
          break;

        case '\f':
          outChars.append("\\f");
          break;

        case '\n':
          outChars.append("\\n");
          break;

        case '\r':
          outChars.append("\\r");
          break;

        case '\t':
          outChars.append("\\t");
          break;

        case (char)0x0b:
          outChars.append("\\v");
          break;

        case '\\':
          outChars.append("\\\\");
          break;

        case '"':
          outChars.append("\\\"");
          break;

        default:
          switch (Character.getType(c)) {
            case Character.CONTROL:
            case Character.PRIVATE_USE:
            case Character.UNASSIGNED:
              if (c <= 0xffff) {
                outChars.append("\\u").append(String.format(Locale.US, "%04X", c));
              }
              else {
                outChars.append("\\U").append(String.format(Locale.US, "%08X", c));
              }

              break;

            default:
              outChars.appendCodePoint(c);
          }
      }

      index += Character.charCount(c);
    }
  }

  @NotNull
  public static GoDecodedString decode(@NotNull String chars) {
    int[] offsets = new int[chars.length() + 1];
    offsets[chars.length()] = -1;

    StringBuilder outChars = new StringBuilder();
    if (chars.indexOf('\\') < 0) {
      outChars.append(chars);
      for (int i = 0; i < offsets.length; i++) {
        offsets[i] = i;
      }
      return new GoDecodedString(true, offsets, outChars.toString());
    }
    int index = 0;

    while (index < chars.length()) {
      char c = chars.charAt(index++);

      offsets[outChars.length()] = index - 1;
      offsets[outChars.length() + 1] = index;

      if (c != '\\') {
        outChars.append(c);
        continue;
      }
      if (index == chars.length()) return new GoDecodedString(false, offsets, outChars.toString());
      c = chars.charAt(index++);

      switch (c) {
        case 'a':
          outChars.append((char)7);
          break;

        case 'b':
          outChars.append('\b');
          break;

        case 'f':
          outChars.append('\f');
          break;

        case 'n':
        case '\n':
          outChars.append('\n');
          break;

        case 'r':
          outChars.append('\r');
          break;

        case 't':
          outChars.append('\t');
          break;

        case 'v':
          outChars.append((char)0x0b);
          break;

        case '\\':
          outChars.append('\\');
          break;

        case '\'':
          outChars.append('\'');
          break;

        case '"':
          outChars.append('"');
          break;


        // octal
        case '0':
        case '1':
        case '2':
        case '3':
        case '4':
        case '5':
        case '6':
        case '7': {
          char startC = c;
          int v = (int)c - '0';
          if (index < chars.length()) {
            c = chars.charAt(index++);
            if ('0' <= c && c <= '7') {
              v <<= 3;
              v += c - '0';
              if (startC <= '3' && index < chars.length()) {
                c = chars.charAt(index++);
                if ('0' <= c && c <= '7') {
                  v <<= 3;
                  v += c - '0';
                }
                else {
                  index--;
                }
              }
            }
            else {
              index--;
            }
          }
          outChars.append((char)v);
        }
        break;

        // hex
        case 'x':
          if (index + 2 <= chars.length()) {
            try {
              int v = Integer.parseInt(chars.substring(index, index + 2), 16);
              outChars.append((char)v);
              index += 2;
            }
            catch (Exception e) {
              return new GoDecodedString(false, offsets, outChars.toString());
            }
          }
          else {
            return new GoDecodedString(false, offsets, outChars.toString());
          }
          break;

        // little unicode
        case 'u':
          if (index + 4 <= chars.length()) {
            try {
              int v = Integer.parseInt(chars.substring(index, index + 4), 16);
              c = chars.charAt(index);
              if (c == '+' || c == '-') return new GoDecodedString(false, offsets, outChars.toString());
              outChars.append((char)v);
              index += 4;
            }
            catch (Exception e) {
              return new GoDecodedString(false, offsets, outChars.toString());
            }
          }
          else {
            return new GoDecodedString(false, offsets, outChars.toString());
          }
          break;

        // big unicode
        case 'U':
          if (index + 8 <= chars.length()) {
            try {
              int v = Integer.parseInt(chars.substring(index, index + 8), 16);
              c = chars.charAt(index);
              if (c == '+' || c == '-') return new GoDecodedString(false, offsets, outChars.toString());
              outChars.append((char)v);
              index += 8;
            }
            catch (Exception e) {
              return new GoDecodedString(false, offsets, outChars.toString());
            }
          }
          else {
            return new GoDecodedString(false, offsets, outChars.toString());
          }
          break;

        default:
          return new GoDecodedString(false, offsets, outChars.toString());
      }

      offsets[outChars.length()] = index;
    }
    return new GoDecodedString(true, offsets, outChars.toString());
  }

  @NotNull
  @Contract(pure = true)
  public static String notNullize(@Nullable String s) {
    return s != null ? s : "";
  }

  @NotNull
  @Contract(pure = true)
  public static String notNullize(@Nullable String s, @NotNull String defaultValue) {
    return s != null ? s : defaultValue;
  }

  @Contract(value = "null -> false", pure = true)
  public static boolean isNotEmpty(@Nullable String s) {
    return s != null && !s.isEmpty();
  }

  @NotNull
  @Contract(pure = true)
  public static String trimStart(@NotNull String s, @NotNull String prefix) {
    if (s.startsWith(prefix)) {
      return s.substring(prefix.length());
    }
    return s;
  }

  @NotNull
  @Contract(pure = true)
  public static String trimEnd(@NotNull String s, @NotNull String suffix) {
    if (s.endsWith(suffix)) {
      return s.substring(0, s.length() - suffix.length());
    }
    return s;
  }

  @NotNull
  @Contract(pure = true)
  public static String trimLeading(@NotNull String string) {
    int index = 0;
    while (index < string.length() && Character.isWhitespace(string.charAt(index))) index++;
    return string.substring(index);
  }

  public static class GoDecodedString {
    public final boolean correct;

    @NotNull public final int[] offsets;
    @NotNull public final String unescapedString;

    public GoDecodedString(boolean correct, @NotNull int[] offsets, @NotNull String unescapedString) {
      this.correct = correct;
      this.offsets = offsets;
      this.unescapedString = unescapedString;
    }
  }

  @Nullable
  @Contract(value = "null -> null", pure = true)
  private static String unquoteStringLiteralExpression(@Nullable String elementText) {
    if (isStringLiteralExpression(elementText)) {
      return elementText.substring(1, elementText.length() - 1);
    }
    return null;
  }

  @Contract(value = "null -> false", pure = true)
  private static boolean isStringLiteralExpression(@Nullable String elementText) {
    if (elementText == null || elementText.length() <= 1) {
      return false;
    }
    char firstChar = elementText.charAt(0);
    return (firstChar == '`' || firstChar == '"') && elementText.charAt(elementText.length() - 1) == firstChar;
  }
}
